using System;
using System.Collections.Generic;
using System.Collections.Immutable;
using System.Linq;
using System.Reflection;
using Upsilon.BaseTypes;
using Upsilon.Binder;
using Upsilon.StandardLibraries;
using Type = Upsilon.BaseTypes.Type;

namespace Upsilon.BoundTypes
{
    public class UserDataBoundTypeDefinition : BoundTypeDefinition
    {
        public string Name { get; protected set; }
        public Dictionary<string, UserDataBoundProperty> Properties { get; protected set; }

        internal UserDataBoundTypeDefinition(System.Type backingType)
            : base(Type.UserData, backingType)
        {
            Name = backingType.Name;
        }

        public UserDataBoundTypeDefinition(string name, Dictionary<string, UserDataBoundProperty> backingType)
            : base(Type.UserData, typeof(void))
        {
            Name = name;
            Properties = backingType;
        }



        public static UserDataBoundTypeDefinition Create(System.Type backingType, string name)
        {
            var obj = new UserDataBoundTypeDefinition(backingType)
            {
                Properties = new Dictionary<string, UserDataBoundProperty>(),
                Name = name
            };
            var fields = backingType.GetFields().Select(x => new UserDataBoundProperty()
            {
                Name       = x.Name,
                ActualType =  x.FieldType.Name,
                Type       = x.FieldType.GetScriptType(),
            });
            foreach (var f in fields)
            {
                obj.Properties.Add(f.Name.ToLowerInvariant(), f);
            }
            var properties = backingType.GetProperties().Select(x => new UserDataBoundProperty()
            {
                Name       = x.Name,
                ActualType =  x.PropertyType.Name,
                Type       = x.PropertyType.GetScriptType(),
            });
            foreach (var f in properties)
            {
                obj.Properties.Add(f.Name.ToLowerInvariant(), f);
            }
            var methods = new List<UserDataBoundMethod>();
            var backingMethods = backingType.GetMethods();
            foreach (var backingMethod in backingMethods)
            {
                if (backingMethod.IsSpecialName)
                    continue;
                var methodName = backingMethod.Name;
                var attribute = backingMethod.GetCustomAttribute(typeof(ScriptFunctionAttribute));
                if (attribute is ScriptFunctionAttribute sfa )
                {
                    methodName = sfa.Name;
                }

                var parameters = backingMethod.GetParameters().Select(parameter => new UserDataBoundFunctionParameter()
                {
                    Name       = parameter.Name,
                    ActualType = parameter.ParameterType.Name,
                    Type       = StaticScope.DeriveValidTypes(parameter.ParameterType),
                    IsOptional = parameter.IsOptional
                }).ToArray();
                methods.Add(new UserDataBoundMethod()
                {
                    Name       = methodName,
                    Type       = Type.Function,
                    ResultType = backingMethod.ReturnType.GetScriptType(),
                    Parameters = parameters
                });
            }
            foreach (var f in methods)
            {
                var cleanedName = f.Name.ToLowerInvariant();
                // TODO: handle this better, considering overloads
                if (obj.Properties.ContainsKey(cleanedName))
                    continue;
                obj.Properties.Add(cleanedName, f);
            }

            return obj;
        }
    }

    public class UserDataBoundEnumDefinition : UserDataBoundTypeDefinition
    {
        public UserDataBoundEnumDefinition(System.Type enumType, string name) : base(enumType)
        {
            if (!enumType.IsEnum)
                throw new Exception("Trying to bind an enum with a type that's not an enum");

            Properties = new Dictionary<string, UserDataBoundProperty>();
            var enumValues         = Enum.GetValues(enumType);

            Name = name;

            for (var i=0; i < enumValues.Length; i++)
            {
                var value = enumValues.GetValue(i);

                var valueName = value.ToString().ToLowerInvariant();
                Properties.Add(valueName, new UserDataBoundProperty()
                {
                    Name = valueName,
                    ActualType = enumType.ToString(),
                    Type = Type.UserData
                });
            }
        }

        public UserDataBoundEnumDefinition(IEnumerable<string> values, string name) : base(name, new Dictionary<string, UserDataBoundProperty>())
        {
            Properties = new Dictionary<string, UserDataBoundProperty>();
            Name = name;

            foreach (var value in values)
            {
                var valueName = value.ToLowerInvariant();
                Properties.Add(valueName, new UserDataBoundProperty()
                {
                    Name = valueName,
                    Type = Type.UserData
                });
            }
        }
    }

    public class UserDataBoundProperty
    {
        public string Name { get; set; }
        public Type Type { get; set; }
        public virtual string ActualType { get; set; }
        public string Comment { get; set; }
    }

    public class UserDataBoundFunctionParameter : UserDataBoundProperty
    {
        public bool IsOptional { get; set; }
    }

    public class UserDataBoundMethod: UserDataBoundProperty
    {
        public override string ActualType => "Function";
        public Type ResultType { get; set; }
        public UserDataBoundFunctionParameter[] Parameters { get; set; }

        public (bool IsValid, string Error,
            BoundExpression WrongParameter) ValidateParameters(ImmutableArray<BoundExpression> callingParameters)
        {
            if (callingParameters.Length   < Parameters.Count(x => !x.IsOptional)
                || callingParameters.Length > Parameters.Length)
            {
                return (false,
                    $"Invalid number of parameters for function '{Name}'. Expected {Parameters.Length}, got {callingParameters.Length}",
                    null);
            }

            for (var i = 0; i < callingParameters.Length; i++)
            {
                var functionParameter = Parameters[i];
                var callingParameter  = callingParameters[i];
                if (callingParameter.Type == Type.Unknown || callingParameter.Type == Type.Nil)
                    continue;

                if (!functionParameter.Type.HasFlag(callingParameter.Type))
                {
                    return (false,
                        $"Unexpected variable passed to internal function at variable {i + 1}. " +
                        $"Expected one of the following: {functionParameter.Type.ToString()}, got: '{callingParameter.Type}'",
                        callingParameter);
                }
            }

            return (true, null, null);
        }

    }

}