using System;
using System.Collections.Generic;
using Upsilon.BaseTypes;
using Upsilon.BaseTypes.Number;
using Upsilon.BaseTypes.ScriptTypeInterfaces;
using Upsilon.Binder;
using Upsilon.Evaluator;

// ReSharper disable UnusedMember.Global
// ReSharper disable MemberCanBePrivate.Global

namespace Upsilon.StandardLibraries
{
    internal class BasicFunctions : ScriptLibrary
    {
        [ScriptFunction("assert", "Asserts that the parameter passed is true. Throws an exception if it is not true. " +
                                                 "Can take a message to show in the exception, otherwise throws with message \"assertion failed!\"",
            directScriptManipulation: true)]
        public void Assert(ScriptBoolean boolean)
        {
            if (!boolean)
            {
                Error(new ScriptString("assertion failed!"));
            }
        }
        public void Assert(ScriptBoolean boolean, ScriptString message)
        {
            if (!boolean)
            {
                Error(message);
            }
        }

        [ScriptFunction("error", "Throw an exception with given error message",
            directScriptManipulation: true)]
        public void Error(ScriptString message)
        {
            throw new Exception(message.Value);
        }

        [ScriptFunction("ipairs", "Iterates over an iterable variable, like a table, until it encounters a nil value.",
            directScriptManipulation: true)]
        public IIterable UpTillNullPairs(IIterable table)
        {
            return new UpTillNullPairsScriptIterator(table);
        }

        [ScriptFunction("pairs", "Iterates over an iterable variable, like a table, skipping all nil values.",
            directScriptManipulation: true)]
        public IIterable Pairs(IIterable table)
        {
            return new PairsScriptIterator(table);
        }

        [ScriptFunction("print", "Prints a message to the action given in the script options", passScriptReference: true,
            directScriptManipulation: true)]
        public void Print(Script script, ScriptType message)
        {
            script.Options.PrintResult.Invoke(message.ToCSharpObject().ToString());
        }

        [ScriptFunction("require", "Loads a module from the module path, using the given script loader.",
            true, true)]
        public ScriptType Require(Script script, ScriptString fileName)
        {
            var file = fileName.Value;
            var module = script.Options.ModuleHandler.GetModule(script, file);
            if (module == null)
            {
                throw new Exception($"Unable to load module {file} at path {script.Options.ScriptLoader.ModulesPath}");
            }
            var s = Executor.ContinueWith(script, module);
            var val = s.Evaluate();
            return val != null ? val.ToScriptType() : new ScriptNull();
        }


        [ScriptFunction("tonumber", "Parses a string to a number.", directScriptManipulation: true)]
        public ScriptNumber ToNumber(ScriptString obj)
        {
            var str = obj.Value;
            if (str.Contains("."))
            {
                return new ScriptNumberDouble(double.Parse(str));
            }
            else
            {
                return new ScriptNumberLong(long.Parse(str));
            }
        }

        [ScriptFunction("tostring", "Returns the string value of the given object.", directScriptManipulation: true)]
        public ScriptString ToString(ScriptType obj)
        {
            return new ScriptString(obj.ToString());
        }

        [ScriptFunction("type", "Returns the type of the given object as a string.", directScriptManipulation: true)]
        public ScriptString Type(ScriptType obj)
        {
            return new ScriptString(obj.Type.ToString());
        }

    }
}